using UnityEngine;
using UnityEngine.UI;
using UnityEngine.XR;
using System;
using System.Collections;
using System.Collections.Generic;
using SimpleJSON;
using System.Linq;
using System.IO;
using MeshVR;
using Request = MeshVR.AssetLoader.AssetBundleFromFileRequest;
using AssetBundles;
using MVR;

// VAMAudioAnalyzer v1.0
//
// Allows precise control of triggers based on audio frequencies bands.
// Based on the excellent tutorial by PeerPlay https://www.youtube.com/watch?v=4Av788P9stk

namespace VAMAudioAnalyzerPlugin
{
    public class VAMAudioAnalyzer : MVRScript
    {
		// JSONStorables
		public JSONStorableFloat FreqIncrement;
		public JSONStorableFloat FreqIncrementChangeSpeed;
		
		private JSONStorableString _helpText;

		// Game Objects and components
		private AudioSource _CurrentAudioSource;
		
		// Variables
		private bool _isSilent = false;
		private static float[] _samples = new float[512];
		private static float[] _freqBand = new float[8];
		
		private static float[] _bandBuffer = new float[8];
		private float[] _bufferDecrease = new float[8];

		private float[] _freqBandHighest = new float[8];
		private static float[] _audioBand = new float[8];
		private static float[] _audioBandBuffer = new float[8];
		
		// Triggers
		private List<EventTrigger> allTriggers;
		private EventTrigger band0Trigger;
		private EventTrigger band1Trigger;
		private EventTrigger band2Trigger;
		private EventTrigger band3Trigger;
		private EventTrigger band4Trigger;
		private EventTrigger band5Trigger;
		private EventTrigger band6Trigger;
		private EventTrigger band7Trigger;
			
		public override void Init()
        {			 
            try
            {
                if (containingAtom.type != "AudioSource" && containingAtom.type != "Empty")
                {
                    SuperController.LogError("Please add VAMAudioAnalyzer on an AudioSource or on an Empty");
                    return;
                }
				
				// Creating all my triggers and adding them to a global list (to easily control them and save them)
				allTriggers = new List<EventTrigger>();

				band0Trigger = new EventTrigger(this, "Low Sub Bass Actions");
				allTriggers.Add(band0Trigger);
				band1Trigger = new EventTrigger(this, "High Sub Bass Actions");
				allTriggers.Add(band1Trigger);
				band2Trigger = new EventTrigger(this, "Low Bass Actions");
				allTriggers.Add(band2Trigger);
				band3Trigger = new EventTrigger(this, "High Bass Actions");
				allTriggers.Add(band3Trigger);
				band4Trigger = new EventTrigger(this, "Low Mids Actions");
				allTriggers.Add(band4Trigger);
				band5Trigger = new EventTrigger(this, "High Mids Actions");
				allTriggers.Add(band5Trigger);
				band6Trigger = new EventTrigger(this, "Low Treble Actions");
				allTriggers.Add(band6Trigger);
				band7Trigger = new EventTrigger(this, "High Treble Actions");
				allTriggers.Add(band7Trigger);
				
				
				// Enabling all triggers
				band0Trigger.active = true;
				band1Trigger.active = true;
				band2Trigger.active = true;
				band3Trigger.active = true;
				band4Trigger.active = true;
				band5Trigger.active = true;
				band6Trigger.active = true;
				band7Trigger.active = true;
				
				// Loading triggers assets
				SimpleTriggerHandler.LoadAssets();
				
				// Creating UI
				
				createStaticDescriptionText("Visualizer values","<color=#000><size=35><b>VISUALIZER VALUES</b></size>\n<size=28>Frequencies values settings</size></color>",false,75);
				
				FreqIncrement = new JSONStorableFloat("Increment", 0.005f, 0.0001f, 0.01f);
				RegisterFloat(FreqIncrement);
				CreateSlider(FreqIncrement).valueFormat = "F4";
				
				FreqIncrementChangeSpeed = new JSONStorableFloat("Increment change speed", 1.2f, 1.001f, 2f);
				RegisterFloat(FreqIncrementChangeSpeed);
				CreateSlider(FreqIncrementChangeSpeed).valueFormat = "F3";
				
				CreateSpacer();

				createStaticDescriptionText("Frequency triggers","<color=#000><size=35><b>FREQUENCIES TRIGGERS</b></size>\n<size=28>Triggers for each band</size></color>",false,75);
				
				UIDynamicButton editLSBT = CreateButton("Low Sub Bass Triggers", false);
				if (editLSBT != null) {
					editLSBT.button.onClick.AddListener( () => {
						band0Trigger.OpenPanelTransition();
					});
				}
				
				UIDynamicButton editHSBT = CreateButton("High Sub Bass Triggers", false);
				if (editHSBT != null) {
					editHSBT.button.onClick.AddListener( () => {
						band1Trigger.OpenPanelTransition();
					});
				}
				
				UIDynamicButton editLBT = CreateButton("Low Bass Triggers", false);
				if (editLBT != null) {
					editLBT.button.onClick.AddListener( () => {
						band2Trigger.OpenPanelTransition();
					});
				}
				
				UIDynamicButton editHBT = CreateButton("High Bass Triggers", false);
				if (editHBT != null) {
					editHBT.button.onClick.AddListener( () => {
						band3Trigger.OpenPanelTransition();
					});
				}
				
				UIDynamicButton editLMT = CreateButton("Low Mids Triggers", false);
				if (editLMT != null) {
					editLMT.button.onClick.AddListener( () => {
						band4Trigger.OpenPanelTransition();
					});
				}
				
				UIDynamicButton editHMT = CreateButton("High Mids Triggers", false);
				if (editHMT != null) {
					editHMT.button.onClick.AddListener( () => {
						band5Trigger.OpenPanelTransition();
					});
				}
				
				UIDynamicButton editLTT = CreateButton("Low Treble Triggers", false);
				if (editLTT != null) {
					editLTT.button.onClick.AddListener( () => {
						band6Trigger.OpenPanelTransition();
					});
				}
				
				UIDynamicButton editHTT = CreateButton("High Treble Triggers", false);
				if (editHTT != null) {
					editHTT.button.onClick.AddListener( () => {
						band7Trigger.OpenPanelTransition();
					});
				}
				
				_helpText = new JSONStorableString("Help",
					"<color=#000><size=35><b>VAMAudioAnalyzer</b></size></color>\n\n" + 
					"<size=32><color=#333>" +
					"This plugin allows you to send dynamic triggers based on an audio source or the scene audio frequencies.\n\nAdd it to an audio source to analyse the sound granularly or add it to an empty atom to analyse the global sound of a scene.\n\n" +
					"<b>Visualizer values</b>\nThese settings will affect the way the 0/1 values will be updated. It will make the update faster, slower or smoother.\n\n" +
					"<b>Frequencies triggers</b>\nEach button allows to edit or create transition actions for each frequency band. They are arranged from Lower to higher (sub bass to treble)." +
					"</color></size>"
				);
				
				UIDynamic HTUID = CreateTextField(_helpText, true);
				HTUID.height = 950.0f;
				
				// Grabbing audio source
				_CurrentAudioSource = containingAtom.reParentObject.GetComponentInChildren(typeof(AudioSource)) as AudioSource;
				if( containingAtom.type == "AudioSource" && _CurrentAudioSource == null ) {
					SuperController.LogError("VAMAudioAnalyzer - Can't find the audio source.");
				}			
				
			}
            catch(Exception e)
            {
                SuperController.LogError("VAMAudioAnalyzer - Exception caught: " + e);
            }
		}
		
		void Start()
        {
			try
            {


			}
            catch(Exception e)
            {
                SuperController.LogError("VAMAudioAnalyzer - Exception caught: " + e);
            }
        }
		
		void Update() {
			// Updating the spectrum itself
			GetSpectrumAudioSource();
			MakeFrequencyBands();
			BandBuffer();
			CreateAudioBands();
			
			// Doing the trigger stuff
			band0Trigger.TriggerTransition( _audioBandBuffer[0] );
			band1Trigger.TriggerTransition( _audioBandBuffer[1] );
			band2Trigger.TriggerTransition( _audioBandBuffer[2] );
			band3Trigger.TriggerTransition( _audioBandBuffer[3] );
			band4Trigger.TriggerTransition( _audioBandBuffer[4] );
			band5Trigger.TriggerTransition( _audioBandBuffer[5] );
			band6Trigger.TriggerTransition( _audioBandBuffer[6] );
			band7Trigger.TriggerTransition( _audioBandBuffer[7] );
		}
		
		// **************************
		// Functions
		// **************************
		private void GetSpectrumAudioSource() {
			if( containingAtom.type == "AudioSource" ) {
				_CurrentAudioSource.GetSpectrumData(_samples, 0, FFTWindow.Blackman);
			} else {
				AudioListener.GetSpectrumData(_samples, 0, FFTWindow.Blackman);
			}
		}
		
		private void CreateAudioBands() {
			for ( int i = 0; i < 8; i++ ) {
				if( _freqBand[i] > _freqBandHighest[i] ) {
					_freqBandHighest[i] = _freqBand[i];
				}
				// If we are silent we return a value of 0 (fix when the track/sound hasn't started)
				_audioBand[i] = _isSilent ? 0 : (_freqBand[i] / _freqBandHighest[i]);
				_audioBandBuffer[i] = _isSilent ? 0 : (_bandBuffer[i] / _freqBandHighest[i]);
			}
		}
		
		private void BandBuffer() {
			for ( int i = 0; i < 8; i++ ) {
				if( _freqBand[i] > _bandBuffer[i] ) {
					_bandBuffer[i] = _freqBand[i];
					_bufferDecrease[i] = FreqIncrement.val;
				}
				
				if( _freqBand[i] < _bandBuffer[i] ) {
					_bandBuffer[i] -= _bufferDecrease[i];
					_bufferDecrease[i] *= FreqIncrementChangeSpeed.val;
				}
			}
		}
		
		private void MakeFrequencyBands() {
			int count = 0;
			float smpSum = 0f;
			for( int i = 0; i < 8; i++ ) {
				
				float average = 0;
				int sampleCount = (int)Mathf.Pow(2,i) * 2;
				if( i == 7 ) {
					sampleCount += 2;
				}
				
				for( int j = 0; j < sampleCount; j++ ) {
					average += _samples[count] * (count + 1);
					count++;
					
				}
				
				average /= count;
				_freqBand[i] = average * 10;
				smpSum += average;
			}
			
			// Fix! When the plugin started initially, the values where always capped at one.
			// I'm using a sum of the samples to evaluate if there is no sound at all
			_isSilent = smpSum <= 0.009f ? true : false; // Arbitraty 0.009 ? Seems to work in my tests, might need some tweaks
		}
		
		// **************************
		// Callbacks & similar
		// **************************
		
		// **************************
		// OVERRIDES FOR THE TRIGGERS
		// **************************	
		public override JSONClass GetJSON(bool includePhysical = true, bool includeAppearance = true, bool forceStore = false)
		{
			JSONClass jc = base.GetJSON(includePhysical, includeAppearance, forceStore);
			if (includePhysical || forceStore)
			{
				needsStore = true;
				foreach( EventTrigger et in allTriggers ) {
					jc[et.Name] = et.GetJSON(base.subScenePrefix);
				}
				
			}
			return jc;
		}

		public override void LateRestoreFromJSON(JSONClass jc, bool restorePhysical = true, bool restoreAppearance = true, bool setMissingToDefault = true)
		{
			base.LateRestoreFromJSON(jc, restorePhysical, restoreAppearance, setMissingToDefault);
			if (!base.physicalLocked && restorePhysical && !IsCustomPhysicalParamLocked("trigger"))
			{
				foreach( EventTrigger et in allTriggers ) {
					et.RestoreFromJSON(jc, base.subScenePrefix, base.mergeRestore, setMissingToDefault);
				}
			}
		}
			
		// **************************
		// Local Tools
		// **************************
		private void logDebug( string debugText ) {
			SuperController.LogMessage( debugText );
		}
		
		private void logError( string debugText ) {
			SuperController.LogError( debugText );
		}
		
		private static void disableScrollOnText(UIDynamicTextField target) {
			ScrollRect targetSR = target.UItext.transform.parent.transform.parent.transform.parent.GetComponent<ScrollRect>();
			if( targetSR != null ) {
				targetSR.horizontal = false;
				targetSR.vertical = false;
			}
		}
		
		private JSONStorableString createStaticDescriptionText( string DescTitle, string DescText, bool rightSide, int fieldHeight ) {
			JSONStorableString staticDescString = new JSONStorableString(DescTitle,DescText);
			UIDynamicTextField staticDescStringField = CreateTextField(staticDescString, rightSide);
			staticDescStringField.backgroundColor = new Color(1f, 1f, 1f, 0f);
			LayoutElement sdsfLayout = staticDescStringField.GetComponent<LayoutElement>();
			sdsfLayout.preferredHeight = sdsfLayout.minHeight = fieldHeight;
			staticDescStringField.height = fieldHeight;
			disableScrollOnText(staticDescStringField);

			return staticDescString;
		}
				
		// **************************
		// Time to cleanup !
		// **************************
		void OnDestroy() {
			// Cleaning modified native VAM assets
			foreach( EventTrigger et in allTriggers ) {
				et.Remove();
			}
		}


		// ***********************************************************
		// EXTERNAL TOOLS - Thank you great coders for your content!
		// ***********************************************************
		
		// *********** MacGruber_Utils.cs START *********************
		
		// ===========================================================================================
		// TriggerHandler implementation for easier handling of custom triggers.
		// Essentially call this in your plugin init code:
		//     StartCoroutine(SimpleTriggerHandler.LoadAssets());
		//
		// Credit to AcidBubbles for figuring out how to do custom triggers.
		//
		// hazmhox note : Edited a bit to allow the use of transitions ( original code coming from LogicBricks 4 )
		public class SimpleTriggerHandler : TriggerHandler
		{
			public static bool Loaded { get; private set; }

			private static SimpleTriggerHandler myInstance;

			private RectTransform myTriggerActionsPrefab;
			private RectTransform myTriggerActionMiniPrefab;
			private RectTransform myTriggerActionDiscretePrefab;
			private RectTransform myTriggerActionTransitionPrefab;

			public static SimpleTriggerHandler Instance {
				get {
					if (myInstance == null)
						myInstance = new SimpleTriggerHandler();
					return myInstance;
				}
			}

			public static void LoadAssets()
			{
				SuperController.singleton.StartCoroutine(Instance.LoadAssetsInternal());
			}

			private IEnumerator LoadAssetsInternal()
			{
				foreach (var x in LoadAsset("z_ui2", "TriggerActionsPanel", p => myTriggerActionsPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionMiniPanel", p => myTriggerActionMiniPrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionDiscretePanel", p => myTriggerActionDiscretePrefab = p))
					yield return x;
				foreach (var x in LoadAsset("z_ui2", "TriggerActionTransitionPanel", p => myTriggerActionTransitionPrefab = p))
					yield return x;

				Loaded = true;
			}

			private IEnumerable LoadAsset(string assetBundleName, string assetName, Action<RectTransform> assign)
			{
				AssetBundleLoadAssetOperation request = AssetBundleManager.LoadAssetAsync(assetBundleName, assetName, typeof(GameObject));
				if (request == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null request.");
				yield return request;
				GameObject go = request.GetAsset<GameObject>();
				if (go == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null GameObject.");
				RectTransform prefab = go.GetComponent<RectTransform>();
				if (prefab == null)
					throw new NullReferenceException($"Request for {assetName} in {assetBundleName} assetbundle failed: Null RectTansform.");
				assign(prefab);
			}


			void TriggerHandler.RemoveTrigger(Trigger t)
			{
				// nothing to do
			}

			void TriggerHandler.DuplicateTrigger(Trigger t)
			{
				throw new NotImplementedException();
			}

			RectTransform TriggerHandler.CreateTriggerActionsUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionsPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionMiniUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionMiniPrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionDiscreteUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionDiscretePrefab);
			}

			RectTransform TriggerHandler.CreateTriggerActionTransitionUI()
			{
				return UnityEngine.Object.Instantiate(myTriggerActionTransitionPrefab);
			}

			void TriggerHandler.RemoveTriggerActionUI(RectTransform rt)
			{
				UnityEngine.Object.Destroy(rt?.gameObject);
			}
		}

		// Wrapper for easier handling of custom triggers.
		public class EventTrigger : Trigger
		{
			public string Name {
				get { return name; }
				set { name = value; myNeedInit = true; }
			}

			public MVRScript Owner {
				get; private set;
			}

			private string name;
			private bool myNeedInit = true;

			public EventTrigger(MVRScript owner, string name)
			{
				Name = name;
				Owner = owner;
				handler = SimpleTriggerHandler.Instance;
			}

			public void OpenPanelActionStart()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab1/Label").GetComponent<Text>().text = "Event Actions";
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
			
			public void OpenPanelTransition()
			{
				if (!SimpleTriggerHandler.Loaded)
				{
					SuperController.LogError("EventTrigger: You need to call SimpleTriggerHandler.LoadAssets() before use.");
					return;
				}

				triggerActionsParent = Owner.UITransform;
				InitTriggerUI();
				OpenTriggerActionsPanel();
				
				//getRoot(triggerActionsParent,0);
				if (myNeedInit)
				{
					triggerActionsPanel.Find("Panel/Header Text").GetComponent<Text>().text = Name;
					triggerActionsPanel.Find("Content/Tab2/Label").GetComponent<Text>().text = "Frequencies Actions";
					triggerActionsPanel.Find("Content/Tab1").gameObject.SetActive(false);
					triggerActionsPanel.Find("Content/Tab2").gameObject.SetActive(true);
					triggerActionsPanel.Find("Content/Tab2").GetComponent<Toggle>().isOn = true;					
					triggerActionsPanel.Find("Content/Tab3").gameObject.SetActive(false);
					myNeedInit = false;
				}
			}
	
			public void RestoreFromJSON(JSONClass jc, string subScenePrefix, bool isMerge, bool setMissingToDefault)
			{
				if (jc.HasKey(Name))
				{
					JSONClass tc = jc[Name].AsObject;
					if (tc != null)
						base.RestoreFromJSON(tc, subScenePrefix, isMerge);
				}
				else if (setMissingToDefault)
				{
					base.RestoreFromJSON(new JSONClass());
				}
			}

			public void Trigger()
			{
				active = true;
				active = false;
			}
			
			public void TriggerTransition(float val)
			{
				this.transitionInterpValue = val;
			}
		}
		// *********** MacGruber_Utils.cs END *********************
	}
}